from pyhocon import ConfigFactory
from pathlib import Path
from jinja2 import Template
import json


class MeltanoAnalysisFileParserError(Exception):
    def __init__(self, message, file_name, *args):
        self.message = message
        self.file_name = file_name
        super(MeltanoAnalysisFileParserError, self).__init__(
            self.message, self.file_name, *args
        )


class MeltanoAnalysisFileParserMissingViewError(MeltanoAnalysisFileParserError):
    def __init__(self, field, your_choice, cls, file_name, *args):
        self.file_name = file_name
        self.cls = cls
        self.message = f'Missing accompanying view "{your_choice}" in "{field}" field in {cls} in {file_name}.'
        super(MeltanoAnalysisFileParserMissingViewError, self).__init__(
            self.message, self.file_name, *args
        )


class MeltanoAnalysisFileParserUnacceptableChoiceError(MeltanoAnalysisFileParserError):
    def __init__(self, field, your_choice, acceptable_choices, cls, file_name, *args):
        self.file_name = file_name
        self.cls = cls
        quoted_choices = [f'"{a}"' for a in acceptable_choices]
        self.message = f"Unacceptable choice for field {field}. You wrote \"{your_choice}\". Acceptable choices are: {', '.join(quoted_choices)}. In {file_name}"
        super(MeltanoAnalysisFileParserUnacceptableChoiceError, self).__init__(
            self.message, self.file_name, *args
        )


class MeltanoAnalysisFileParserMissingFieldsError(MeltanoAnalysisFileParserError):
    def __init__(self, fields, cls, file_name, *args):
        self.file_name = file_name
        self.cls = cls
        field = "field" if len(fields) == 1 else "fields"
        self.message = (
            f"Missing {field} in {cls}: \"{', '.join(fields)}\" in {self.file_name}"
        )
        super(MeltanoAnalysisFileParserMissingFieldsError, self).__init__(
            self.message, self.file_name, *args
        )

    def __str__(self):
        return f"{self.message} in {self.file_name}"

    def __repr__(self):
        return f"{self.message} in {self.file_name}"


class MeltanoAnalysisFileParser:
    def __init__(self, directory):
        self.directory = directory
        self.models = []
        self.required_model_properties = ["name", "connection", "label", "explores"]
        self.required_explore_properties = ["from", "label", "description"]
        self.required_join_properties = ["sql_on", "relationship"]
        self.required_view_properties = ["sql_table_name", "dimensions"]
        self.join_relationship_types = [
            "one_to_one",
            "one_to_many",
            "many_to_one",
            "many_to_many",
        ]

    def uses_accepted_choice(self, choices, choice):
        return choice in choices

    def missing_properties(self, properties, properties_dict):
        properties_copy = properties.copy()
        for prop in properties_dict:
            try:
                property_index = properties_copy.index(prop)
                del properties_copy[property_index]
            except ValueError as e:
                continue
        return properties_copy

    def parse_ma_file(self, file_path):
        try:
            return ConfigFactory.parse_string(open(file_path, "r").read())
        except Exception as e:
            raise MeltanoAnalysisFileParserError(str(e), str(file_path.parts[-1]))

    def compile(self, models):
        indices = {}
        for model in models:
            compiled_file_name = f"{model['name']}.model.mac"
            compiled_file_path = Path(self.directory).joinpath(compiled_file_name)
            compiled_model = open(compiled_file_path, "w")
            indices[model["name"]] = {
                "explores": [e["name"] for e in model["explores"]]
            }
            compiled_model.write(json.dumps(model))
            compiled_model.close()

        # index file
        index_file_path = Path(self.directory).joinpath("models.index.mac")
        index_file = open(index_file_path, "w")
        index_file.write(json.dumps(indices))
        index_file.close()

    def parse(self):
        self.ma_views = Path(self.directory).glob("*.view.ma")
        self.ma_models = Path(self.directory).glob("*.model.ma")
        self.ma_dashboards = Path(self.directory).glob("*.dashboards.ma")
        for model in self.ma_models:
            file_name = model.parts[-1]
            conf = self.parse_ma_file(model)
            parsed_model = self.model(conf, file_name)
            self.models.append(parsed_model)
        return self.models

    def model(self, ma_file_model_dict, file_name):
        temp_model = {}
        missing_properties = self.missing_properties(
            self.required_model_properties, ma_file_model_dict
        )
        if missing_properties:
            raise MeltanoAnalysisFileParserMissingFieldsError(
                missing_properties, "model", file_name
            )
        for prop_name, prop_def in ma_file_model_dict.items():
            temp_model[prop_name] = prop_def
            if prop_name == "explores":
                temp_model[prop_name] = self.explores(prop_def, file_name)
        return temp_model

    def view_conf_by_name(self, view_name, cls, prop, file_name):
        try:
            return next(
                view
                for view in self.ma_views
                if view.parts[-1] == f"{view_name}.view.ma"
            )
        except StopIteration as e:
            raise MeltanoAnalysisFileParserMissingViewError(
                prop, view_name, cls, file_name
            )

    def explores(self, ma_file_explores_dict, file_name):
        model_explores = []
        for explore_name, explore_def in ma_file_explores_dict.items():
            temp_explore = {}
            temp_explore["name"] = explore_name
            missing_properties = self.missing_properties(
                self.required_explore_properties, explore_def
            )
            if missing_properties:
                raise MeltanoAnalysisFileParserMissingFieldsError(
                    missing_properties, "explore", file_name
                )
            for prop_name, prop_def in explore_def.items():
                temp_explore[prop_name] = prop_def
                if prop_name == "from":
                    matching_view = self.view_conf_by_name(
                        temp_explore[prop_name], "explore", prop_name, file_name
                    )
                    temp_explore["related_view"] = self.view(
                        self.parse_ma_file(matching_view), matching_view.parts[-1]
                    )
                if prop_name == "joins":
                    temp_explore[prop_name] = self.joins(prop_def, file_name)
            model_explores.append(temp_explore)
        return model_explores

    def joins(self, ma_file_joins_dict, file_name):
        explore_joins = []
        for join_name, join_def in ma_file_joins_dict.items():
            temp_join = {}
            temp_join["name"] = join_name
            matching_view = self.view_conf_by_name(
                temp_join["name"], "join", "name", file_name
            )
            temp_join["related_view"] = self.view(
                self.parse_ma_file(matching_view), matching_view.parts[-1]
            )
            missing_properties = self.missing_properties(
                self.required_join_properties, join_def
            )
            if missing_properties:
                raise MeltanoAnalysisFileParserMissingFieldsError(
                    missing_properties, "join", file_name
                )
            for prop_name, prop_def in join_def.items():
                temp_join[prop_name] = prop_def
                if prop_name == "relationship":
                    uses_accepted_choices = self.uses_accepted_choice(
                        self.join_relationship_types, prop_def
                    )
                    if not uses_accepted_choices:
                        raise MeltanoAnalysisFileParserUnacceptableChoiceError(
                            prop_name,
                            prop_def,
                            self.join_relationship_types,
                            "join",
                            file_name,
                        )

            explore_joins.append(temp_join)
        return explore_joins

    def view(self, view_file, file_name):
        temp_view = {}
        missing_properties = self.missing_properties(
            self.required_view_properties, view_file
        )
        if missing_properties:
            raise MeltanoAnalysisFileParserMissingFieldsError(
                missing_properties, "view", file_name
            )
        for prop_name, prop_def in view_file.items():
            temp_view[prop_name] = prop_def
            if prop_name == "dimensions":
                temp_view[prop_name] = self.dimensions(prop_def)
            elif prop_name == "measures":
                temp_view[prop_name] = self.measures(prop_def)
        return temp_view

    def dimensions(self, ma_file_dimensions_dict):
        temp_dimensions = []
        for dimension_name, dimension_def in ma_file_dimensions_dict.items():
            temp_dimension = {"name": dimension_name}
            for prop_name, prop_def in dimension_def.items():
                temp_dimension[prop_name] = prop_def
            temp_dimensions.append(temp_dimension)
        return temp_dimensions

    def measures(self, ma_file_measures_dict):
        temp_measures = []
        for measure_name, measure_def in ma_file_measures_dict.items():
            temp_measure = {"name": measure_name}
            for prop_name, prop_def in measure_def.items():
                temp_measure[prop_name] = prop_def
            temp_measures.append(temp_measure)
        return temp_measures



import os
import click
import logging
import fnmatch
import json

from . import cli
from meltano.core.project import Project
from meltano.core.config_service import ConfigService
from meltano.core.plugin import Plugin, PluginType
from meltano.core.plugin.error import PluginExecutionError
from meltano.core.plugin_invoker import PluginInvoker
from meltano.core.plugin.singer.catalog import (
    visit,
    parse_select_pattern,
    ListSelectedExecutor,
)


@cli.command()
@click.argument("extractor")
@click.argument("entities_filter", default="*")
@click.argument("attributes_filter", default="*")
@click.option("--list", is_flag=True)
@click.option("--all", is_flag=True)
@click.option("--exclude", is_flag=True)
def select(extractor, entities_filter, attributes_filter, **flags):
    try:
        if flags["list"]:
            show(extractor, entities_filter, attributes_filter, show_all=flags["all"])
        else:
            add(extractor, entities_filter, attributes_filter, exclude=flags["exclude"])
    except PluginExecutionError as e:
        raise click.ClickException(
            f"Cannot list the selected properties: "
            "there was a problem running the tap with `--discover`. "
            "Make sure the tap supports `--discover` and run "
            "`meltano invoke {extractor} --discover` to make "
            "sure it runs correctly."
        ) from e
    except Exception as e:
        raise click.ClickException(str(e)) from e


def add(extractor, entities_filter, attributes_filter, exclude=False):
    exclude = "!" if exclude else ""
    project = Project.find()
    config = ConfigService(project)
    pattern = f"{exclude}{entities_filter}.{attributes_filter}"

    with project.meltano_update() as meltano:
        extractor = config.get_plugin(PluginType.EXTRACTORS, extractor)
        extractor.add_select_filter(pattern)

        idx = next(i for i, it in enumerate(config.get_extractors()) if it == extractor)
        meltano["extractors"][idx] = extractor.canonical()


def show(extractor, entities_filter, attributes_filter, show_all=False):
    project = Project.find()
    config = ConfigService(project)
    extractor = config.get_plugin(PluginType.EXTRACTORS, extractor)
    invoker = PluginInvoker(project, extractor)
    pattern = f"{entities_filter}.{attributes_filter}"

    list_all = ListSelectedExecutor()
    try:
        if not invoker.files["catalog"].exists():
            logging.info("Catalog not found, trying to run the tap with --discover.")
            extractor.run_discovery(invoker)

        with invoker.files["catalog"].open() as catalog:
            schema = json.load(catalog)
            visit(schema, list_all)
    except FileNotFoundError as e:
        logging.error(
            "Cannot find catalog: make sure the tap runs correctly with --discover; `meltano invoke TAP --discover`"
        )
        raise e
    else:
        invoker.files["catalog"].unlink()

    # report
    click.secho("Enabled patterns:")
    for select in map(parse_select_pattern, extractor.select):
        click.secho(
            f"\t{select.property_pattern}", fg="red" if select.negated else "white"
        )
    else:
        click.echo()

    click.secho("Selected properties:")
    color = lambda selected: "white" if selected else "red"
    for stream, prop in (
        (stream, prop)
        for stream, props in list_all.properties.items()
        for prop in props
    ):
        if show_all:
            click.secho(f"\t{stream.key}", fg=color(stream.selected), nl=False)
            click.echo(".", nl=False)
            click.secho(prop.key, fg=color(stream.selected and prop.selected))
        elif stream.selected and prop.selected:
            click.echo(f"\t{stream.key}.{prop.key}")



import os
from dotenv import load_dotenv

load_dotenv()

LOG_PATH = os.getenv("MELTANO_LOG_PATH", "meltano.log")
ENV = "development"

API_ROOT_DIR = os.path.abspath(os.path.dirname(__file__))
TEMP_FOLDER = os.path.join(API_ROOT_DIR, "static/tmp")
PROJECT_ROOT_DIR = os.path.dirname(API_ROOT_DIR)

# SQLALCHEMY_ECHO=True

THREADS_PER_PAGE = 2

SECRET_KEY = "damnitjanice"



import os
import logging
import click
import datetime

from . import cli
from .params import db_options
from meltano.core.runner.singer import SingerRunner
from meltano.core.runner.dbt import DbtRunner
from meltano.core.project import Project, ProjectNotFound
from meltano.core.plugin import PluginType


@cli.command()
@db_options
@click.argument("extractor")
@click.argument("loader")
@click.option("--dry", help="Do not actually run.", is_flag=True)
@click.option("--transform", type=click.Choice(["skip", "only", "run"]), default="skip")
@click.option(
    "--job_id", envvar="MELTANO_JOB_ID", help="A custom string to identify the job."
)
def elt(extractor, loader, dry, transform, job_id):
    """
    meltano elt EXTRACTOR_NAME LOADER_NAME

    extractor_name: Which extractor should be used in this extraction
    loader_name: Which loader should be used in this extraction
    """
    try:
        project = Project.find()
    except ProjectNotFound as e:
        raise click.ClickException(e)

    if job_id is None:
        # Autogenerate a job_id if it is not provided by the user
        job_id = f'job_{datetime.datetime.now().strftime("%Y%m%d-%H:%M:%S.%f")}'

    singer_runner = SingerRunner(
        project,
        job_id=job_id,
        run_dir=os.getenv("SINGER_RUN_DIR", project.meltano_dir("run")),
        target_config_dir=project.meltano_dir(PluginType.LOADERS, loader),
        tap_config_dir=project.meltano_dir(PluginType.EXTRACTORS, extractor),
    )

    dbt_runner = DbtRunner(project)

    try:
        if transform != "only":
            click.echo("Running extract & load...")
            singer_runner.perform(extractor, loader, dry_run=dry)
            click.secho("Extract & load complete!", fg="green")
        else:
            click.secho("Extract & load skipped.", fg="yellow")

        if transform != "skip":
            click.echo("Running transformation...")
            dbt_runner.perform(dry_run=dry, models=extractor)
            click.secho("Transformation complete!", fg="green")
        else:
            click.secho("Transformation skipped.", fg="yellow")
    except Exception as err:
        raise click.ClickException(
            f"ELT could not complete, an error happened during the process: {err}."
        )



import json

from typing import Dict
from pathlib import Path


class ExploreMissingError(Exception):
    pass


class ConnectionNotFoundError(Exception):
    pass


class MACFile:
    def __init__(self, content: Dict):
        self.content = content

    @classmethod
    def load(cls, file):
        return MACFile(json.load(file))

    @property
    def explores(self):
        return self.content["explores"]

    def explore(self, explore_name: str) -> Dict:
        try:
            return next(e for e in self.explores if e["name"] == explore_name)
        except StopIteration:
            raise ExploreMissingError(f"{explore_name} not found.")

    def connection(self, connection_name: str) -> Dict:
        try:
            return self.content["connection"]
        except StopIteration:
            raise ConnectionNotFoundError(f"{connection_name} not found.")



import os
import yaml
import logging
from pathlib import Path
from typing import Union, Dict
from contextlib import contextmanager
from functools import wraps

from .plugin import Plugin
from .error import Error


class ProjectNotFound(Error):
    def __init__(self):
        super().__init__("Cannot find `meltano.yml`. Are you in a meltano project?")


class Project:
    """
    Represent the current Meltano project from a file-system
    perspective.
    """

    _meltano = {}

    def __init__(self, root: Union[Path, str]):
        self.root = Path(root)

    def chdir(self):
        os.chdir(self.root)
        self.root = Path(".")

    @classmethod
    def find(self, from_dir: Union[Path, str] = ".", chdir=True):
        """
        Recursively search for a `meltano.yml` file. Once found,
        return a `Project` correct dir.
        """
        cwd = os.getcwd()

        while not Project(".").meltanofile.exists():
            if os.getcwd() == "/":
                raise ProjectNotFound()

            os.chdir("..")

        if not chdir:
            os.chdir(cwd)

        return Project(os.getcwd())

    @property
    def meltano(self) -> Dict:
        """Return a copy of the current meltano config"""
        if not self._meltano:
            self._meltano = yaml.load(self.meltanofile.open()) or {}
        return self._meltano.copy()

    @contextmanager
    def meltano_update(self):
        """
        Yield the current meltano configuration and update the meltanofile
        if the context ends gracefully.
        """
        try:
            meltano_update = self.meltano
            yield meltano_update

            # save it
            with self.meltanofile.open("w") as meltanofile:
                meltanofile.write(yaml.dump(meltano_update, default_flow_style=False))

            # update the cache
            self._meltano = meltano_update
        except Exception as err:
            logging.error(f"Could not update meltano.yml: {err}")
            raise

    def makedirs(func):
        @wraps(func)
        def decorate(*args, **kwargs):
            path = func(*args, **kwargs)

            # if there is an extension, only create the base dir
            _, ext = os.path.splitext(path)
            if ext:
                dir = os.path.dirname(path)
            else:
                dir = path

            os.makedirs(dir, exist_ok=True)
            return path

        return decorate

    @property
    def meltanofile(self):
        return self.root.joinpath("meltano.yml")

    @makedirs
    def meltano_dir(self, *joinpaths):
        return self.root.joinpath(".meltano", *joinpaths)

    @makedirs
    def venvs_dir(self, *prefixes):
        return self.meltano_dir(*prefixes, "venv")

    @makedirs
    def run_dir(self, *joinpaths):
        return self.meltano_dir("run", *joinpaths)

    @makedirs
    def plugin_dir(self, plugin: Plugin, *joinpaths):
        return self.meltano_dir(plugin.type, plugin.name, *joinpaths)



import subprocess
import logging

from .venv_service import VenvService
from .plugin import PluginType


class DbtService:
    def __init__(self, project, venv_service: VenvService = None):
        self.project = project
        self.venv_service = venv_service or VenvService(project)
        self.transform_dir = f"{self.project.root}/transform/"
        self.profile_dir = f"{self.project.root}/transform/profile/"

    @property
    def exec_path(self):
        return self.venv_service.exec_path("dbt", namespace=PluginType.TRANSFORMERS)

    def call(self, *args):
        logging.debug(f"Invoking: dbt {args}")
        exec_args = list(map(str, [self.exec_path, *args]))
        run = subprocess.run(exec_args, cwd=self.transform_dir)

        run.check_returncode()
        return run

    def compile(self, models=None):
        params = ["compile", "--profiles-dir", self.profile_dir, "--profile", "meltano"]
        if models:
            params.extend(["--models", models])

        return self.call(*params)

    def deps(self):
        return self.call("deps", "--profiles-dir", self.profile_dir)

    def run(self, models=None):
        params = ["run", "--profiles-dir", self.profile_dir, "--profile", "meltano"]
        if models:
            params.extend(["--models", models])

        return self.call(*params)



import datetime
import logging
import os
from logging.handlers import RotatingFileHandler
from flask import Flask, request, render_template
from flask import jsonify
from flask_cors import CORS
from jinja2.exceptions import TemplateNotFound

from . import config
from .external_connector import ExternalConnector

app = Flask(__name__)

app.config.from_object(config)

flask_env = os.getenv("FLASK_ENV", "development")

if flask_env == "development":
    CORS(app)

connector = ExternalConnector()

logger = logging.getLogger("melt_logger")
handler = RotatingFileHandler(app.config["LOG_PATH"], maxBytes=2000, backupCount=10)
logger.addHandler(handler)
now = str(datetime.datetime.utcnow().strftime("%b %d %Y %I:%M:%S:%f"))
logger.warning(f"Melt started at: {now}")


@app.before_request
def before_request():
    logger.info(f"[{request.remote_addr}] request: {now}")


@app.errorhandler(500)
def internal_error(exception):
    logger.info(f"[{request.remote_addr}] request: {now}, error: {exception}")
    return jsonify({"error": 1}), 500


@app.route("/model")
@app.route("/")
def analyze():
    try:
        return render_template("analyze.html")
    except TemplateNotFound:
        return "Please run `yarn build` from src/analyze."


@app.route("/drop")
def drop_it():
    from .controllers.sqlhelper import SqlHelper

    SqlHelper().reset_db()
    return jsonify({"dropped_it": "like it's hot"})


from .controllers.repos import reposBP
from .controllers.settings import settingsBP
from .controllers.sql import sqlBP

# from .controllers.orchestrations import orchestrationsBP

app.register_blueprint(reposBP)
app.register_blueprint(settingsBP)
app.register_blueprint(sqlBP)
# app.register_blueprint(orchestrationsBP)



import json
import os
from collections import OrderedDict
from datetime import date, datetime
from decimal import Decimal
from pathlib import Path
from os.path import join

from flask import Blueprint, jsonify, request
import sqlalchemy

from .sqlhelper import SqlHelper
from .settingshelper import SettingsHelper
from meltano.core.mac_file import MACFile
from meltano.core.project import Project

sqlBP = Blueprint("sql", __name__, url_prefix="/sql")
meltano_model_path = join(os.getcwd(), "model")


class ConnectionNotFound(Exception):
    def __init__(self, connection_name: str):
        self.connection_name = connection_name
        super().__init__("{connection_name} is missing.")


@sqlBP.errorhandler(ConnectionNotFound)
def _handle(ex):
    return (
        jsonify(
            {
                "error": True,
                "code": f"Missing connection details to '{ex.connection_name}'. Create a connection to '{ex.connection_name}' in the settings.",
            }
        ),
        500,
    )


@sqlBP.errorhandler(sqlalchemy.exc.DBAPIError)
def _handle(ex):
    return (
        jsonify(
            {"error": True, "code": ex.code, "orig": str(ex), "statement": ex.statement}
        ),
        422,
    )


def default(obj):
    if isinstance(obj, Decimal):
        return str(obj)
    elif isinstance(obj, (datetime, date)):
        return obj.isoformat()
    raise TypeError(f"Object of type {type(obj).__name__} is not JSON serialize-able")


def get_db_engine(connection_name):
    project = Project.find()
    settings_helper = SettingsHelper()
    connections = settings_helper.get_connections()["settings"]["connections"]

    try:
        connection = next(
            connection
            for connection in connections
            if connection["name"] == connection_name
        )

        if connection["dialect"] == "postgresql":
            psql_params = ["username", "password", "host", "port", "database"]
            user, pw, host, port, db = [connection[param] for param in psql_params]
            connection_url = f"postgresql+psycopg2://{user}:{pw}@{host}:{port}/{db}"
        elif connection["dialect"] == "sqlite":
            db_path = project.root.joinpath(connection["path"])
            connection_url = f"sqlite:///{db_path}"

        return sqlalchemy.create_engine(connection_url)

        raise ConnectionNotFound(connection_name)
    except StopIteration:
        raise ConnectionNotFound(connection_name)


@sqlBP.route("/", methods=["GET"])
def index():
    return jsonify({"result": True})


@sqlBP.route("/get/<model_name>/<explore_name>", methods=["POST"])
def get_sql(model_name, explore_name):
    mac_file = Path(meltano_model_path).joinpath(f"{model_name}.model.mac")
    with mac_file.open() as f:
        mac = MACFile.load(f)

    explore = mac.explore(explore_name)
    incoming_json = request.get_json()

    to_run = incoming_json["run"]
    incoming_order = incoming_json["order"]

    sqlHelper = SqlHelper()
    sql_dict = sqlHelper.get_sql(explore, incoming_json)
    outgoing_sql = sql_dict["sql"]
    measures = sql_dict["measures"]
    dimensions = sql_dict["dimensions"]
    column_headers = sql_dict["column_headers"]
    names = sql_dict["names"]

    if not to_run:
        return json.dumps({"sql": outgoing_sql}, default=default)

    connection_name = mac.connection("connection")
    engine = get_db_engine(connection_name)
    results = engine.execute(outgoing_sql)

    results = [OrderedDict(row) for row in results]
    base_dict = {"sql": outgoing_sql, "results": results, "error": False}
    if not len(results):
        base_dict["empty"] = True
    else:
        base_dict["empty"] = False
        base_dict["column_headers"] = column_headers
        base_dict["names"] = names
        base_dict["keys"] = list(results[0].keys())
        base_dict["measures"] = sqlHelper.get_names(measures)

    return json.dumps(base_dict, default=default)


@sqlBP.route("/distinct/<model_name>/<explore_name>", methods=["POST"])
def get_distinct_field_name(model_name, explore_name):
    # incoming_json = request.get_json()
    # field_name = incoming_json["field"].replace("${TABLE}", explore_name)
    # model = Model.query.filter(Model.name == model_name).first()
    # explore = Explore.query.filter(Explore.name == explore_name).first()

    # base_table = explore.view.settings["sql_table_name"]
    # base_sql = f"SELECT DISTINCT {field_name} FROM {base_table} AS {explore_name} ORDER BY {field_name}"

    # engine = get_db_engine(model.settings["connection"])
    # results = engine.execute(base_sql)
    # results = [dict(row) for row in results]
    # return json.dumps(
    #     {"sql": base_sql, "results": results, "keys": list(results[0].keys())},
    #     default=default,
    # )
    return json.dumps({"tester": "tester"})



from typing import Dict
from meltano.core.plugin import PluginType

from .catalog import visit, SelectExecutor, ListExecutor
from .base import SingerPlugin
from .tap import SingerTap
from .target import SingerTarget


def plugin_factory(plugin_type: PluginType, canonical: Dict):
    plugin_class = {PluginType.EXTRACTORS: SingerTap, PluginType.LOADERS: SingerTarget}

    # this will parse the discovery file and create an instance of the
    # corresponding `plugin_class` for all the plugins.
    return plugin_class[plugin_type](**canonical)



import logging
import re

import sqlalchemy
from flask import jsonify
from pypika import Query, Order

from .analysishelper import AnalysisHelper
from .date import Date
from .joinhelper import JoinHelper


class SqlHelper:
    def parse_sql(self, input):
        placeholders = self.placeholder_match(input)

    def placeholder_match(self, input):
        outer_pattern = r"(\$\{[\w\.]*\})"
        inner_pattern = r"\$\{([\w\.]*)\}"
        outer_results = re.findall(outer_pattern, input)
        inner_results = re.findall(inner_pattern, input)
        return (outer_results, inner_results)

    def get_names(self, things):
        return [thing["name"] for thing in things]

    def get_sql(self, explore, incoming_json):
        view_name = incoming_json["view"]
        view = explore["related_view"]

        base_table = view["sql_table_name"]
        incoming_dimensions = incoming_json["dimensions"]
        incoming_dimension_groups = incoming_json["dimension_groups"]
        incoming_measures = incoming_json["measures"]
        incoming_filters = incoming_json["filters"]
        incoming_joins = incoming_json["joins"]
        incoming_limit = incoming_json.get("limit", 50)
        incoming_order = incoming_json["order"]
        order = None
        if incoming_order:
            if incoming_order["direction"] == "asc":
                order = Order.asc
            else:
                order = Order.desc
        orderby = incoming_order["column"] if incoming_order else None

        # get all timeframes
        timeframes = [t["timeframes"] for t in incoming_dimension_groups]
        # flatten list of timeframes
        timeframes = [y for x in timeframes for y in x]
        dimensions = AnalysisHelper.dimensions_from_names(incoming_dimensions, view)
        measures = AnalysisHelper.measures_from_names(incoming_measures, view)
        dimensions_raw = dimensions
        measures_raw = measures

        table = AnalysisHelper.table(base_table, explore["name"])
        joins = [JoinHelper.get_join(j) for j in incoming_joins]
        dimension_groups = self.dimension_groups(
            view_name, incoming_dimension_groups, table
        )
        dimensions = AnalysisHelper.dimensions(dimensions, table)
        dimensions = dimensions + dimension_groups
        measures = AnalysisHelper.measures(measures, table)

        if orderby:
            ordered_by_column = [d for d in dimensions_raw if d.name == orderby]
            if ordered_by_column:
                orderby = self.dimensions(ordered_by_column, table)[0]
            else:
                ordered_by_column = [m for m in measures_raw if m.name == orderby]
                if ordered_by_column:
                    orderby = self.measures(ordered_by_column, table)[0]
                else:
                    raise Exception(
                        "Something is wrong, no dimension or measure column matching the column to sort by."
                    )

        column_headers = self.column_headers(dimensions_raw, measures_raw)
        names = self.get_names(dimensions_raw + measures_raw)
        return {
            "dimensions": dimensions_raw,
            "measures": measures_raw,
            "column_headers": column_headers,
            "names": names,
            "sql": self.get_query(
                from_=table,
                dimensions=dimensions,
                measures=measures,
                limit=incoming_limit,
                joins=joins,
                order=order,
                orderby=orderby,
            ),
        }

    def column_headers(self, dimensions, measures):
        return [d["label"] for d in dimensions + measures]

    def dimension_groups(self, view_name, dimension_groups, table):
        fields = []
        for dimension_group in dimension_groups:
            dimension_group_queried = (
                DimensionGroup.query.join(View, DimensionGroup.view_id == View.id)
                .filter(View.name == view_name)
                .filter(DimensionGroup.name == dimension_group["name"])
                .first()
            )
            (_table, name) = dimension_group_queried.table_column_name.split(".")
            for timeframe in dimension_group["timeframes"]:
                d = Date(timeframe, table, name)
                fields.append(d.sql)
        return fields

    def get_query(
        self, from_, dimensions, measures, limit, joins=None, order=None, orderby=None
    ):
        select = dimensions + measures
        q = Query.from_(from_)
        if joins:
            region = from_
            entry = joins[0]["table"]
            join_dimensions = joins[0]["dimensions"]
            join_measures = joins[0]["measures"]
            select = select + join_dimensions + join_measures
            dimensions = dimensions + join_dimensions
            q = q.join(joins[0]["table"]).on(
                region.id == entry.region_id and region.id == entry.region_id
            )
        q = q.select(*select).groupby(*dimensions)
        if order:
            q = q.orderby(orderby, order=order)
        q = q.limit(limit)
        return f"{str(q)};"

    def reset_db(self):
        try:
            Settings.__table__.drop(db.engine)
            Project.__table__.drop(db.engine)
            db.drop_all()
        except sqlalchemy.exc.OperationalError as err:
            logging.error("Failed drop database.")
        db.create_all()
        settings = Settings()
        db.session.add(settings)
        db.session.commit()
        return jsonify({"dropped_it": "like_its_hot"})



from flask import Blueprint, jsonify, request
from .settingshelper import SettingsHelper

settingsBP = Blueprint("settings", __name__, url_prefix="/settings")


@settingsBP.route("/", methods=["GET"])
def index():
    settings_helper = SettingsHelper()
    return jsonify(settings_helper.get_connections())


@settingsBP.route("/save", methods=["POST"])
def save():
    settings_helper = SettingsHelper()
    connection = request.get_json()
    settings = settings_helper.save_connection(connection)
    return jsonify(settings)


@settingsBP.route("/delete", methods=["POST"])
def delete():
    settings_helper = SettingsHelper()
    connection = request.get_json()
    settings = settings_helper.delete_connection(connection)
    return jsonify(settings)


@settingsBP.route("/connections/<name>/test")
def test(name):
    current_settings = Settings.query.first().settings
    connections = current_settings["connections"]
    try:
        found_connection = next(
            connection for connection in connections if connection["name"] == name
        )
    # this is a really broad exception catch, this will swallow sneaky errors
    except Exception as e:
        found_connection = {}

    return jsonify(found_connection)



from enum import Enum

from .analysishelper import AnalysisHelper


class JoinType(Enum):
    left_join = "left"
    inner_join = "inner"
    full_outer_join = "full_outer"
    cross_join = "cross"


class JoinHelper:
    @staticmethod
    def get_join(join):
        # view = JoinHelper.get_view(join)
        # table = AnalysisHelper.table(view.settings["sql_table_name"], view.name)
        # dimensions = JoinHelper.get_dimensions(join, view, table)
        # # TODO finish this implementation with measures
        # measures = []
        # join = Join.query.filter(Join.name == join["name"]).first()
        # return {
        #     "view": view,
        #     "dimensions": dimensions,
        #     "measures": measures,
        #     "table": table,
        #     "on": join.settings["sql_on"],
        #     "join": join,
        # }
        return {"get_join": "get_join"}

    def get_dimensions(join, view, table):
        dimensions = AnalysisHelper.dimensions_from_names(join["dimensions"], view)
        return AnalysisHelper.dimensions(dimensions, table)

    def get_view(join):
        # return View.query.filter(View.name == join["name"]).first()
        return {}



import yaml
import fnmatch
from collections import namedtuple
from enum import Enum


class YAMLEnum(str, Enum):
    def __str__(self):
        return self.value

    @staticmethod
    def yaml_representer(dumper, obj):
        return dumper.represent_scalar("tag:yaml.org,2002:str", str(obj))


yaml.add_multi_representer(YAMLEnum, YAMLEnum.yaml_representer)


class PluginType(YAMLEnum):
    EXTRACTORS = "extractors"
    LOADERS = "loaders"
    TRANSFORMERS = "transformers"
    ALL = "all"

    def __str__(self):
        return self.value


class Plugin:
    def __init__(
        self,
        plugin_type: PluginType,
        name: str,
        pip_url=None,
        config=None,
        select=None,
        **extras
    ):
        self.name = name
        self.type = plugin_type
        self.pip_url = pip_url
        self.config = config
        self._select = set(select or [])
        self._extras = extras or {}

    def canonical(self):
        canonical = {
            "name": self.name,
            "pip_url": self.pip_url,
            "config": self.config,
            **self._extras,
        }

        if self._select:
            canonical.update({"select": list(self._select)})

        if self._extras:
            canonical.update(**self._extras)

        return canonical

    @property
    def config_files(self):
        """Return a list of stubbed files created for this plugin."""
        return []

    @property
    def output_files(self):
        return []

    @property
    def select(self):
        return self._select or {"*.*"}

    @select.setter
    def select(self, patterns):
        self._select = set(patterns)

    def add_select_filter(self, filter: str):
        self._select.add(filter)

    def __eq__(self, other):
        return self.name == other.name and self.type == other.type



import os
import subprocess
import click

from . import cli
from meltano.api import app


@cli.command()
@click.option("--debug/--no-debug", default=True, help="Run webserver in debug mode")
@click.option("--port", default=5000, help="Port to run webserver on")
@click.option(
    "--reload/--no-reload",
    default=True,
    help="To reload the server or not on file changes",
)
@click.option("--hostname", default="0.0.0.0", help="The hostname of the webserver")
def ui(debug, port, reload, hostname):
    # todo: run gunicorn if not in debug mode
    if debug:
        app.run(debug=debug, use_reloader=reload, port=port, host=hostname)
    else:
        pass



import logging
import re
from fnmatch import fnmatch
from collections import OrderedDict, namedtuple
from enum import Enum, auto
from functools import singledispatch
from typing import List


SelectPattern = namedtuple(
    "SelectPattern", ("stream_pattern", "property_pattern", "negated")
)


def parse_select_pattern(pattern: str):
    negated = False

    if pattern.startswith("!"):
        negated = True
        pattern = pattern[1:]

    stream, *_ = pattern.split(".")

    return SelectPattern(
        stream_pattern=stream, property_pattern=pattern, negated=negated
    )


class CatalogNode(Enum):
    STREAM = auto()
    STREAM_METADATA = auto()
    STREAM_PROPERTY = auto()
    STREAM_PROPERTY_METADATA = auto()


class CatalogExecutor:
    def execute(self, node_type: CatalogNode, node, path):
        dispatch = {
            CatalogNode.STREAM: self.stream_node,
            CatalogNode.STREAM_METADATA: self.stream_metadata_node,
            CatalogNode.STREAM_PROPERTY_METADATA: self.property_metadata_node,
            CatalogNode.STREAM_PROPERTY: self.property_node,
        }

        try:
            dispatch[node_type](node, path)
        except KeyError:
            logging.debug(f"Unknown node type '{node_type}'.")

    def stream_node(self, node, path: str):
        pass

    def property_node(self, node, path: str):
        pass

    def stream_metadata_node(self, node, path: str):
        pass

    def property_metadata_node(self, node, path: str):
        pass

    def __call__(self, node_type, node, path):
        return self.execute(node_type, node, path)


class SelectExecutor(CatalogExecutor):
    def __init__(self, patterns: List[str]):
        self._stream = None
        self._patterns = list(map(parse_select_pattern, patterns))

    @classmethod
    def _match_patterns(cls, value, include=[], exclude=[]):
        included = any(fnmatch(value, pattern) for pattern in include)
        excluded = any(fnmatch(value, pattern) for pattern in exclude)

        return included and not excluded

    def stream_match_patterns(self, stream):
        return self._match_patterns(
            stream,
            include=(
                pattern.stream_pattern
                for pattern in self._patterns
                if not pattern.negated
            ),
        )

    def property_match_patterns(self, prop):
        return self._match_patterns(
            prop,
            include=(
                pattern.property_pattern
                for pattern in self._patterns
                if not pattern.negated
            ),
            exclude=(
                pattern.property_pattern
                for pattern in self._patterns
                if pattern.negated
            ),
        )

    def stream_node(self, node, path):
        self._stream = node["stream"]

        if not self.stream_match_patterns(self._stream):
            logging.debug(f"{self._stream} is not selected.")
            return

        found = any(
            metadata
            for metadata in node["metadata"]
            if len(metadata["breadcrumb"]) == 0
        )

        # This is to support legacy catalogs
        if not found:
            node["metadata"].insert(
                0,
                {
                    "breadcrumb": [],
                    "metadata": {"inclusion": "available", "selected": True},
                },
            )
            logging.debug(f"{path} has been selected.")

        node.update({"selected": True})

    def stream_metadata_node(self, node, path):
        if not self.stream_match_patterns(self._stream):
            logging.debug(f"{self._stream} is not selected.")
            return

        metadata = node["metadata"]
        metadata.update({"selected": True})
        logging.debug(f"{path} has been selected.")

    def property_metadata_node(self, node, path):
        _, name = node["breadcrumb"]
        prop = f"{self._stream}.{name}"

        if not self.property_match_patterns(prop):
            logging.debug(f"{prop} is not selected.")
            return

        metadata = node["metadata"]
        if metadata.get("inclusion") == "available":
            metadata.update({"selected": True})
            logging.debug(f"{path} has been selected.")


class ListExecutor(CatalogExecutor):
    def __init__(self, selected_only=False):
        # properties per stream
        self._selected_only = selected_only
        self.properties = OrderedDict()

        super().__init__()

    def stream_node(self, node, path):
        stream = node["stream"]
        if stream not in self.properties:
            self.properties[stream] = set()

    def property_node(self, node, path):
        *_, name = path.split(".")
        # current stream
        stream = next(reversed(self.properties))
        self.properties[stream].add(name)


class ListSelectedExecutor(CatalogExecutor):
    SelectedNode = namedtuple("SelectedNode", ("key", "selected"))

    def __init__(self):
        self.properties = OrderedDict()
        super().__init__()

    @property
    def selected_properties(self):
        selected = {}
        for stream, props in self.properties.items():
            if not stream.selected:
                continue

            selected[stream.key] = set()
            for prop in filter(lambda node: node.selected, props):
                selected[stream.key].add(prop.key)

        return selected

    def is_node_selected(self, node):
        try:
            metadata = node["metadata"]
            return metadata.get("inclusion") == "automatic" or metadata.get("selected")
        except KeyError:
            return False

    def stream_node(self, node, path):
        self._stream = node["stream"]

    def stream_metadata_node(self, node, path):
        stream = self.SelectedNode(self._stream, self.is_node_selected(node))
        if stream not in self.properties:
            self.properties[stream] = set()

    def property_metadata_node(self, node, path):
        *_, name = node["breadcrumb"]
        property = self.SelectedNode(name, self.is_node_selected(node))

        # current stream
        stream = next(reversed(self.properties))
        self.properties[stream].add(property)


@singledispatch
def visit(node, executor, path: str = ""):
    logging.debug(f"Skipping node at '{path}'")


@visit.register(dict)
def _(node: dict, executor, path=""):
    logging.debug(f"Visiting node at '{path}'.")
    if re.search(r"streams\[\d+\]$", path):
        executor(CatalogNode.STREAM, node, path)

    if re.search(r"schema\.properties\..*$", path):
        executor(CatalogNode.STREAM_PROPERTY, node, path)

    if re.search(r"metadata\[\d+\]$", path) and "breadcrumb" in node:
        if len(node["breadcrumb"]) == 0:
            executor(CatalogNode.STREAM_METADATA, node, path)
        else:
            executor(CatalogNode.STREAM_PROPERTY_METADATA, node, path)

    for child_path, child_node in node.items():
        visit(child_node, executor, path=f"{path}.{child_path}")


@visit.register(list)
def _(node: list, executor, path=""):
    for index, child_node in enumerate(node):
        visit(child_node, executor, path=f"{path}[{index}]")



import base64
import json
import os
import subprocess
import sys
from pathlib import Path
from os.path import join

import markdown
import pkg_resources
from flask import Blueprint, jsonify

from .ma_file_parser import MeltanoAnalysisFileParser, MeltanoAnalysisFileParserError

reposBP = Blueprint("repos", __name__, url_prefix="/repos")
meltano_model_path = join(os.getcwd(), "model")


@reposBP.route("/", methods=["GET"])
def index():
    # For all you know, the first argument to Repo is a path to the repository
    # you want to work with
    onlyfiles = [
        f
        for f in os.listdir(meltano_model_path)
        if os.path.isfile(os.path.join(meltano_model_path, f))
    ]
    sortedMaFiles = {"documents": [], "views": [], "models": [], "dashboards": []}
    onlydocs = Path(meltano_model_path).parent.glob("*.md")
    for d in onlydocs:
        file_dict = {"path": str(d), "abs": str(d), "visual": str(d.name)}
        file_dict["unique"] = base64.b32encode(bytes(file_dict["abs"], "utf-8")).decode(
            "utf-8"
        )
        sortedMaFiles["documents"].append(file_dict)

    for f in onlyfiles:
        filename, ext = os.path.splitext(f)
        if ext != ".ma":
            continue
        file_dict = {"path": f, "abs": f, "visual": f}
        file_dict["unique"] = base64.b32encode(bytes(file_dict["abs"], "utf-8")).decode(
            "utf-8"
        )
        filename = filename.lower()

        filename, ext = os.path.splitext(filename)
        file_dict["visual"] = filename
        if ext == ".view":
            sortedMaFiles["views"].append(file_dict)
        if ext == ".model":
            sortedMaFiles["models"].append(file_dict)
        if ext == ".dashboard":
            sortedMaFiles["dashboards"].append(file_dict)

    return jsonify(sortedMaFiles)


@reposBP.route("/file/<unique>", methods=["GET"])
def file(unique):
    file_path = base64.b32decode(unique).decode("utf-8")
    (filename, ext) = os.path.splitext(file_path)
    is_markdown = False
    path_to_file = os.path.abspath(os.path.join(meltano_model_path, file_path))
    with open(path_to_file, "r") as read_file:
        data = read_file.read()
        if ext == ".md":
            data = markdown.markdown(data)
            is_markdown = True
        return jsonify(
            {
                "file": data,
                "is_markdown": is_markdown,
                "unique": unique,
                "populated": True,
            }
        )


def lint_all(compile):
    ma_parse = MeltanoAnalysisFileParser(meltano_model_path)
    models = ma_parse.parse()
    if compile:
        ma_parse.compile(models)
    return jsonify({"result": True})


@reposBP.errorhandler(MeltanoAnalysisFileParserError)
def handle_meltano_analysis_file_parser_error(e):
    return jsonify(
        {"result": False, "errors": [{"message": e.message, "file_name": e.file_name}]}
    )


@reposBP.errorhandler(FileNotFoundError)
def handle_file_not_found(e):
    return jsonify({"result": False, "error": str(e)})


@reposBP.route("/lint", methods=["GET"])
def lint():
    return lint_all(False)


@reposBP.route("/sync", methods=["GET"])
def sync():
    return lint_all(True)


@reposBP.route("/test", methods=["GET"])
def db_test():
    explore = Explore.query.first()
    return jsonify({"explore": {"name": explore.name, "settings": explore.settings}})


@reposBP.route("/models", methods=["GET"])
def models():
    models = Path(meltano_model_path).joinpath("models.index.mac")
    return jsonify(json.loads(open(models, "r").read()))


@reposBP.route("/explores", methods=["GET"])
def explores():
    explores = Explore.query.all()
    explores_json = []
    for explore in explores:
        explores_json.append(explore.serializable())
    return jsonify(explores_json)


@reposBP.route("/views/<view_name>", methods=["GET"])
def view_read(view_name):
    file_path = Path(meltano_model_path).joinpath(f"{view_name}.view.ma")
    ma_parse = MeltanoAnalysisFileParser(meltano_model_path)
    view = ma_parse.parse_ma_file(file_path)
    return jsonify(view)


@reposBP.route("/explores/<model_name>/<explore_name>", methods=["GET"])
def explore_read(model_name, explore_name):
    model = Path(meltano_model_path).joinpath(f"{model_name}.model.mac")
    with model.open() as f:
        model = json.load(f)
    explores = model["explores"]
    explore = next(e for e in explores if e["from"] == explore_name)
    return jsonify(explore)
